# $Id: Repsys.pm 270568 2010-08-16 18:41:45Z guillomovitch $

package MDV::Repsys;

use strict;
use warnings;
use Carp;
use Cwd;
use Config::IniFiles;
use Date::Parse;
use Date::Format;
use Exception::Class (
    'Mdvsys::Exception'
);
use File::Temp qw(tempdir);
use File::Path;
use File::Spec;
use POSIX qw(setlocale LC_ALL);
use RPM;
use RPM::Header;
use RPM::Spec;

use SVN::Client;
use SVN::Core;
use UNIVERSAL::require;
use URI;
use version; our $VERSION = qv('2.2.0');

use constant {
    DEFAULT_VERSION  => 'current',
    DEFAULT_REVISION => 'HEAD',
    TEMPDIR_TEMPLATE => 'mdvsys_XXXX'
};

=head1 NAME

MDV::Repsys

=head1 DESCRIPTION

Perl interface with Mandriva build system

=head1 AUTHORS

Olivier Thauvin <nanardon@mandriva.org>

Guillaume Rousse <guillomovitch@mandriva.org>

Michaël Scherer <misc@mandriva.org>

=cut

my %rpm_dirs = (
    sourcedir => 'SOURCES',
    patchdir  => 'SOURCES',
    specdir   => 'SPECS',
    builddir  => 'BUILD',
    rpmdir    => 'RPMS',
    srcrpmdir => 'SRPMS',
);

setlocale(LC_ALL, "C");

=head1 CLASS METHODS

=head2 new(%options)

Create a new MDV::Repsys object

Available options:

=over

=item configfile

Use this file as system configuration file

=item build true/false

Allow build action (default: true)

=item commit true/false

Allow commit action (default: true)

=item submit true/false

Allow submit action (default: true).

=back

=cut

sub new {
    my $class = shift;
    my %options = (
        verbosity  => 0,
        build      => 1,
        commit     => 1,
        submit     => 1,
        @_
    );

    my $config = Config::IniFiles->new(
        -file => $options{configfile}
    );
    croak "Unable to read configuration file $options{configfile}\n" 
        unless $config;

    # override system configuration with local configuration if present
    my $user_configfile = "$ENV{HOME}/.mdvsys.conf";
    if (-r $user_configfile) {
        $config = Config::IniFiles->new(
            -file   => $user_configfile,
            -import => $config
        );
    }

    my $binrepos = $config->val('repository', 'split_binaries');
    my $mirror   = $config->val('repository', 'mirror');

    my $self = bless {
        config    => $config,
        binrepos  => (defined $binrepos && $binrepos eq 'yes'),
        mirror    => (defined $mirror),
        svn       => SVN::Client->new(),
        build     => $options{build},
        commit    => $options{commit},
        submit    => $options{submit} && $options{commit} ,
        verbosity => $options{verbosity},
    }, $class;

    return $self;
}

=head1 INSTANCE METHODS

=head2 set_verbosity($level)

Set verbosity to given level.

=cut

sub set_verbosity {
    my ($self, $level) = @_;

    $self->{verbosity} = $level;
}

=head2 unpack_package($rpmfile, $destdir)

Unpack a source package into a directory.
Return corresponding spec file.

=cut

sub unpack_package {
    my ($self, $rpmfile, $destdir) = @_;

    $self->_print_msg(3, 'Unpacking package %s', $rpmfile);

    if (! -d $destdir) {
        mkdir $destdir or croak "Can't create $destdir: $!";
    }

    $self->_set_rpm_dirs_from_topdir($destdir);

    # make rpm output informations starting from log level 4
    RPM::setverbosity($self->{verbosity} >= 4 ? $self->{verbosity} + 2 : 0);
    my ($specfile) = RPM::installsrpm($rpmfile);
    RPM::setverbosity(0);

    return $specfile;
}

=head2 find_current_sources($dir)

Return the list of sources (meaning: sources, icons and spec files) from the
package found in given directory, according to their usage in the spec file.

=cut

sub find_current_sources {
    my ($self, $workdir) = @_;

    $self->_set_rpm_dirs_from_topdir($workdir);
    my $specfile = (glob(RPM::expand_macro('%_specdir/*.spec')))[0]
        or croak "Can't find any spec file\n";
    $self->_print_msg(3, 'Looking sources from spec file %s', $specfile);
   
    my $spec = RPM::Spec->new($specfile, anyarch => 1, force => 0)
        or croak "Can't parse spec file $specfile\n";

    my @sources = ( $specfile ); # specfile should be first !
    my $sourcedir = RPM::expand_macro('%_sourcedir');
    foreach my $file ($spec->sources(), $spec->icon()) {
        push(@sources, "$sourcedir/$file");
    }

    return @sources;
}

=head2 find_unsync_sources($dir, @sources)

Return two lists of files that should be added or removed from the svn
repository according to their current status.

=cut

sub find_unsync_sources {
    my ($self, $workdir, @sources) = @_;

    # remap sources as seen by rpm as actual source files
    my %sources =
        map { $_ => 1 }
        map { -l $_ ? readlink($_) : $_ }
        @sources;

    my @dirs = qw/SPECS SOURCES/;

    # ensure spec and source directory are already versionned
    my $children = $self->{svn}->ls($workdir, 'HEAD', 0);
    foreach my $dir (@dirs) {
        next if $children->{$dir};
        $self->_print_msg(4, "Adding %s", $dir);
        $self->{svn}->add("$workdir/$dir", 0);
    }

    push(@dirs, 'SOURCES-bin') if $self->{binrepos};

    my (@needadd, @needdel);
    foreach my $dir (@dirs) {
        $self->{svn}->status(
            "$workdir/$dir",
            'HEAD',
            sub {
                my ($entry, $status) = @_;
                # skip directory itself
                return if $entry eq "$workdir/$dir";

                my $text_status = $status->text_status();
                if ($text_status == $SVN::Wc::Status::unversioned) {
                    push(@needadd, $entry) if $sources{$entry};
                } elsif (
                    $text_status == $SVN::Wc::Status::normal   ||
                    $text_status == $SVN::Wc::Status::added    ||
                    $text_status == $SVN::Wc::Status::modified
                ) {
                    push(@needdel, $entry) if !$sources{$entry};
                }
            },
            1, # recursive
            1, # get_all
            0, # update
            1, # no_ignore
        );
    }
    
    return(\@needadd, \@needdel);
}

=head2 sync_sources($needadd, $needdel)

Add and remove sources from the repository.

=cut

sub sync_sources {
    my ($self, $needadd, $needdel) = @_;

    foreach my $toadd (sort @{$needadd || []}) {
        $self->_print_msg(4, "Adding %s", $toadd);
        $self->{svn}->add($toadd, 0);
    }
    foreach my $todel (sort @{$needdel || []}) {
        $self->_print_msg(4, "Removing %s", $todel);
        $self->{svn}->delete($todel, 1);
        if ($todel =~ 'SOURCES-bin') {
            # delete corresponding symlink
            my $link = $todel;
            $todel =~ s{SOURCES-bin}{SOURCES};
            unlink $todel;
        }
    }
}

=head2 split_sources($txt_source_dir, $bin_source_dir, %options)

Split binary and text sources in two distinct directories.

=cut

sub split_sources {
    my ($self, $txt_source_dir, $bin_source_dir, %options) = @_;

    return unless $self->{binrepos};

    # move sources files from text repository work directory
    # to binary repository one
    foreach my $file (glob("$txt_source_dir/*")) {
        next if -l $file;
        next if $file !~ /\.(tar\.(gz|bz2)|tgz|zip)$/;
        my $new_file = $file;
        $new_file =~ s{$txt_source_dir}{$bin_source_dir};
        rename $file, $new_file
            or croak "Can't rename $file to $new_file: $!\n";
        if ($options{link}) {
            symlink $new_file, $file
                or croak "Can't link $new_file to $file: $!\n";
        }
    }
}

=head2 build_package($specfile, %options)

Build package from given spec file. Return the source package, followed by the
list of binary packages.

Available options:

=over

=item stage $stage

run rpmbuild to stage $stage

=item binaries_destdir $dir

create binary packages in given directory.

=item source_destdir $dir

create source package in given directory.

=item define => values

Pass each given value to underlying command, using --define switch.

=back

=cut

sub build_package {
    my ($self, $specfile, %options) = @_;

    return unless $self->{build};

    $self->_print_msg(3, 'Building package from spec file %s', $specfile);

    # compute top directory
    $specfile = File::Spec->rel2abs($specfile);
    my ($topdir) = $specfile =~ m{^(.*)/[^/]+/[^/]+\.spec$};

    my @results;
    # set a command to install build dependencies if present in configuration
    # and not explicitely disabled
    my $command =
        (! defined $options{install} || $options{install}) ?
        $self->{config}->val('helper', 'install_buildrequires') || undef :
        undef;

    Youri::Package::RPM::Builder->require(qv('0.2.0')) or croak $@;
    my $builder = Youri::Package::RPM::Builder->new(
        topdir                 => $topdir,
        build_requires_command => $command,
        build_results_callback => sub { @results = @_ },
        verbose                =>
            ($self->{verbosity} > 2 ? $self->{verbosity} - 2 : 0)
    );

    my $options = " --define '_signature \"\"'";
    $options .= " --define '_rpmdir $options{binaries_destdir}'"
        if $options{binaries_destdir};
    $options .= " --define '_srcrpmdir $options{source_destdir}'"
        if $options{source_destdir};
    $options .= " --short-circuit"
        if $options{'short-circuit'};
    $options .= " --define '$_'" foreach
        $options{define} ? @{$options{define}} : ();

    eval {
        $builder->build(
            $specfile,
            stage   => $options{stage},
            options => $options
        );
    }; 
    if ($@) {
        $self->_print_msg(2, 'Keeping build directory in %s', $topdir)
            if $options{'keep-on-failure'};
        croak "Can't build package: $@\n";
    }

    return @results;
}

=head2 get_packages_top_url()

Return the top-level svn url for packages

=cut

sub get_packages_top_url {
    my ($self, %options) = @_;

    my $url;
    my $distribution = $options{distribution} || 'cooker';
    if ($options{binary}) {
        if ($self->{binrepos}) {
            $url = $self->{config}->val('repository', $distribution);
        } else {
            $url = $self->{config}->val('repository', 'main');
            $url .= $self->_get_distribution_path($distribution);
        }
    } else {
        # return preferentialy mirror url, if defined and possible
        if ($self->{mirror} && !$options{write}) {
            $url = $self->{config}->val('repository', 'mirror');
        } else {
            $url = $self->{config}->val('repository', 'main');
        }
        $url .= $self->_get_distribution_path($distribution);
    }

    return $url;
}

sub _get_distribution_path {
    my ($self, $distribution) = @_;

    return $distribution eq 'cooker' ?
        '/cooker' : "/updates/$distribution";
}


=head2 get_package_url($package, %options)

Return the svn url for package named $package

=cut

sub get_package_url {
    my ($self, $package, %options) = @_;
    return sprintf(
        '%s/%s',
        $self->get_package_parent_url($package, %options),
        $options{version} || DEFAULT_VERSION
    );
}

=head2 get_package_parent_url($package, %options)

Return the svn parent url for package named $package

=cut

sub get_package_parent_url {
    my ($self, $package, %options) = @_;

    my $distribution = $options{distribution} || 'cooker';
    my $branch       = $options{branch};
    my $url =
        $self->get_packages_top_url(%options)                              .
        "/$package"                                                        .
        ($branch ? "/branches/$branch" : '');

    return $url;
}

=head2 get_package_name_from_current_directory

Return the package name from current directory

=cut

sub get_package_name_from_current_directory { 
    my ($self) = @_;

    my $url;
    my $receiver = sub {
        my (undef, $info, undef) = @_;
        $url = URI->new($info->URL());
    };
    eval { 
        $self->{svn}->info(getcwd(), undef, 'WORKING', $receiver, 0);
    };
    if ($@) {
        return;
    }

    # split path in a directory list
    my @path = File::Spec->splitdir($url->path());

    # look for a known pattern in the list
    for my $i (0 .. $#path) {
        if ($path[$i] eq 'cooker') {
            return $path[$i + 1];
        }
        if ($path[$i] eq 'updates') {
            return $path[$i + 2];
        }
    }

    return;
}

=head2 checkout_package($package, $destdir, %options)

Checkout a package from svn into $destdir directory

Available options:

=over

=item revision => $revision

Work on a specific revision.

=item distribution => $distribution

Work on a specific distribution tree.

=item branch => $branch

Work on a specific package branch inside a distribution.

=back

=cut

sub checkout_package {
    my ($self, $package, $destdir, %options) = @_;

    $destdir ||= $package;

    $self->_print_msg(3,
        'Checking out package %s into directory %s', $package, $destdir
    );

    my $revision;

    # check out text repository first
    eval {
        my $url = $self->get_package_url($package, %options);
        $revision = $self->{svn}->checkout(
            $url,
            $destdir,
            $options{revision} || DEFAULT_REVISION,
            1,
        );
    };
    if ($@) {
        my $error = ($@ =~ /^Bad URL passed to RA layer/) ?
            "No such package $package in text repository" :
            "Can't checkout $package from text repository: $@\n";
        Mdvsys::Exception->throw($error);
    }

    if ($self->{binrepos}) {
        # check out binary repository second if needed
        eval {
            my $url = $self->get_package_url($package, %options, binary => 1);
            $revision = $self->{svn}->checkout(
                $url . '/SOURCES',
                $destdir . '/SOURCES-bin',
                $options{revision} || DEFAULT_REVISION,
                1,
            );
        };
        if ($@) {
            my $error = ($@ =~ /^Bad URL passed to RA layer/) ?
                "No such package $package in binary repository" :
                "Can't checkout $package from binary repository: $@";
            Mdvsys::Exception->throw($error);
        }

        # symlink files from SOURCES-bin to SOURCES
        my $old_dir = getcwd();
        chdir "$destdir/SOURCES";
        foreach my $file (glob("../SOURCES-bin/*")) {
            my $link = $file;
            $link =~ s{../SOURCES-bin/}{};
            symlink $file, $link or croak "Can't link $file to $link: $!\n";
        }
        chdir $old_dir;
    }

    return $revision;
}

=head2 print_raw_package_changelog($package, $handle, %options)

Print the raw subversion log for package to $handle.
If not specified, $handle is set to STDOUT.

=cut

sub print_raw_package_changelog {
    my ($self, $package, $handle, %options) = @_;

    $handle ||= \*STDOUT;

    $self->_print_msg(3, 'Printing raw changelog for package %s', $package);

    foreach my $entry ($self->get_new_changelog($package, %options)) {
        print $handle $self->_format_changelog_entry($entry);
    }
}

=head2 print_changelog($package, $handle, %options)

Print the rpm-formated log for package to $handle.
If not specified, $handle is set to STDOUT.

Available options:

=over

=item revision => $revision

Work on a specific revision.

=item distribution => $distribution

Work on a specific distribution tree.

=item branch => $branch

Work on a specific package branch inside a distribution.

=back

=cut

sub print_changelog {
    my ($self, $package, $handle, %options) = @_;

    $handle ||= \*STDOUT;

    $self->_print_msg(3, 'Printing changelog for package %s', $package);
    my @entries = (
        $self->get_new_changelog($package, %options),
        $self->get_old_changelog($package, %options)
    );
 
    print $handle "\%changelog\n";

    foreach my $entry (sort {
            $b->{time} &&  $a->{time} ?
            $b->{time} <=> $a->{time} :
            $a->{time} <=> $b->{time}
        } grep { $_ } @entries) {
        print $handle $self->_format_changelog_entry($entry);
    }
}

=head2 strip_changelog_from_spec($specfile)

Available options:

=over

=item target => $filename

write new spec file as given filename

=back

=cut

sub strip_changelog_from_spec {
    my ($self, $specfile, %options) = @_;
    
    $self->_print_msg(3, 'Stripping changelog from spec file %s', $specfile);

    my $new_specfile = $options{target} ? $options{target} : $specfile . '.new';
    open(my $in, '<', $specfile)
        or croak "Can't open $specfile for reading: $!";
    open(my $out, '>', $new_specfile)
        or croak "Can't open $new_specfile for writing: $!";

    while (my $line = <$in>) {
        last if $line =~ /^\%changelog/;
        print $out $line;
    }

    my $changelog;
    while (my $line = <$in>) {
        next if !$line && !$changelog;
        $changelog .= $line;
    }

    close($out);
    close($in);

    if (!$options{target}) {
        rename($new_specfile, $specfile)
            or croak "Can't move $new_specfile to $specfile: $!";
    };

    return $changelog;
}

=head2 add_changelog_to_spec($specfile, %options)

Rewrite a spec file, adding changelog from repository.

Available options:

=over

=item target => $filename

write new spec file as given filename

=back

=cut

sub add_changelog_to_spec {
    my ($self, $specfile, %options) = @_;

    $self->_print_msg(3, 'Adding changelog to spec file %s', $specfile);

    $self->_set_rpm_dirs_from_specfile($specfile);
    my $spec = RPM::Spec->new($specfile, anyarch => 1, force => 1)
        or croak "Can't parse specfile $specfile\n";
    my $header = $spec->srcheader();
    my $package = $header->queryformat('%{NAME}');

    my $new_specfile = $options{target} ? $options{target} : $specfile . '.new';
    open(my $in, '<', $specfile)
        or croak "Can't open $specfile for reading: $!";
    open(my $out, '>', $new_specfile)
        or croak "Can't open $new_specfile for writing: $!";

     while (my $line = <$in>) {
        print $out $line;
    }

    $self->print_changelog(
        $package,
        $out,
        %options,
    );

    close($out);
    close($in);

    if (!$options{target}) {
        rename($new_specfile, $specfile)
            or croak "Can't move $new_specfile to $specfile: $!";
    };

}

=head2 get_last_package_revision($package, %options)

Return the last revision change for a package.

=cut

sub get_last_package_revision {
    my ($self, $package, %options) = @_;
    my $url = $self->get_package_url($package, %options);
    my $leafs;

    eval {
        $leafs = $self->{svn}->ls(
            $url, 
            $options{revision} || DEFAULT_REVISION,
            1,
        );
    };
    if ($@) {
        croak "Can't get information from $url: $@\n";
    }
    my $revision = 0;
    foreach my $leaf (%{$leafs || {}}) {
        next unless defined $leafs->{$leaf};
        if ($leafs->{$leaf}->created_rev > $revision) {
            $revision = $leafs->{$leaf}->created_rev;
        }
    }
        
    return $revision;
}

=head2 get_last_directory_revision($dir, %options)

Return the real last revision change for package checkout into $dir.

=cut

sub get_last_directory_revision {
    my ($self, $dir, %options) = @_;

    $self->_print_msg(4, 'Finding last revision from directory %s', $dir);
    my $revision = 0;
    eval {
        $self->{svn}->status(
            $dir,
            $options{revision} || DEFAULT_REVISION,
            sub {
                my ($path, $status) = @_;
                my $entry = $status->entry();
                return unless $entry;
                my $entry_revision = $entry->cmt_rev();
                $revision = $entry_revision if $revision < $entry_revision;
            },
            1, # recursive
            1, # get_all
            0, # update
            0, # no_ignore
        );
    };
    if ($@) {
        croak "can't get status of $dir: $@\n";
    }

    return $revision
}

=head2 extract_package($package, %options)

Extract a source package from the subversion repository into the current
directory. Return the package file.

Available options:

=over

=item revision => $revision

Work on a specific revision.

=item distribution => $distribution

Work on a specific distribution tree.

=item branch => $branch

Work on a specific package branch inside a distribution.

=item destdir => $dir

Use given directory as destination.

=item prefix => true/false

Prefix the package file by revision number, so as to be compatible with
mandriva submission tool.

=back

=cut

sub extract_package {
    my ($self, $package, %options) = @_;

    $self->_print_msg(2, 'Extracting package %s', $package);

    my $dir = tempdir(
        TEMPDIR_TEMPLATE,
        TMPDIR => 1,
        CLEANUP => ($options{'keep-on-failure'} ? 0 : 1)
    );

    $self->checkout_package($package, $dir, %options);

    my $specfile = "$dir/SPECS/$package.spec";
    
    $self->add_changelog_to_spec(
        $specfile,
        %options,
        package => $package
    );

    my $destdir = $options{destdir} || getcwd();
    my ($file) = $self->build_package(
        $specfile,
        deps           => 0,
        binaries       => 0,
        source_destdir => $destdir
    );

    if ($options{prefix}) {
        my $revision = $self->get_last_directory_revision($dir, %options);
        my $new_file = '@' . $revision . (File::Spec->splitpath($file))[2];
        $self->_print_msg(3, 'Renaming %s to %s', $file, $new_file);
        rename $file, $new_file;
        $file = $new_file;
    }

    # remove temp directory immediatly, not at program termination
    rmtree($dir);

    return $file;
}

=head2 create_package_directory($package, %options)

Create a package directory on the svn.

=over

=item message => $message

Use given commit message (default: "create X", where X is package name).

=back

=cut

sub create_package_directory {
    my ($self, $package, %options) = @_;

    my $parent_url = $self->get_package_parent_url($package, %options);
    my $url = $self->get_package_url($package, %options);

    if ($self->_check_url_exists($url, %options)) {
        croak "$package is already inside svn\n";
    }

    $self->_print_msg(3, "Creating directory for package %s", $package);

    my @urls = (
        $parent_url,
        $url,
        "$url/SOURCES",
        "$url/SPECS",
    );

    if ($options{branch}) {
        my $branch_url = $self->get_package_parent_url(
            $package,
            %options, 
            branch => undef
        ) . '/branches';
        push(@urls, $branch_url)
            unless $self->_check_url_exists($branch_url);
    }
   
    my $message = $options{message} || "create $package";
    $self->{svn}->log_msg(sub { $_[0] = \$message; return 0; });
    eval { $self->{svn}->mkdir(\@urls); };
    $self->{svn}->log_msg(undef);
    if ($@) {
        croak "Can't create $package: $@\n";
    }
}

=head2 import_package($rpmfile, %options)

Import a source package into the svn.

=over

=item message => $message

Use given commit message (default: "import X", where X is package name)

=item submit => true/false

Submit package just after being imported

=back

=cut

sub import_package {
    my ($self, $rpmfile, %options) = @_;

    my $header = rpm2header($rpmfile)
        or croak "Can't read rpm file $rpmfile\n";
    if ($header->hastag('SOURCERPM')) {
        croak "$rpmfile is not a source package\n";
    }
    my $package = $header->queryformat('%{NAME}');

    croak "$package is already inside text repository\n"
        if $self->_check_url_exists($self->get_package_url($package, %options));

    if ($self->{binrepos}) {
        croak "$package is already inside binary repository\n"
            if $self->_check_url_exists($self->get_package_url($package, %options, binary => 1));
    }

    $self->_print_msg(2, "Importing package %s", $package);
    my $dir = tempdir(TEMPDIR_TEMPLATE, TMPDIR => 1, CLEANUP => 1);

    my $txtrepo_url = $self->get_packages_top_url(%options);
    my $txtrepo_dir = $dir . '/txtrepo';
    eval {
        $self->{svn}->checkout(
            $txtrepo_url,
            $txtrepo_dir,
            'HEAD',
            0, # no recursion
        );
    };
    if ($@) {
        croak "Can't checkout $txtrepo_url: $@\n";
    }

    my $txtrepo_pkgdir = $txtrepo_dir . "/$package";
    mkdir $txtrepo_pkgdir;

    my $specfile = $self->unpack_package(
        $rpmfile,
        "$txtrepo_pkgdir/current",
    );

    my $changelog = $self->strip_changelog_from_spec(
        $specfile, 
        %options,
        package => $package,
    );

    if ($changelog) {
        my $split_changelog =
            $self->{config}->val('repository', 'split_changelog');
        if ($split_changelog && $split_changelog eq 'yes') {
            # manage changelog separatly
            my $logrepo_url =
                $self->{config}->val('repository', 'main') .
                '/misc';
            my $logrepo_dir = $dir . '/log';
            eval {
                $self->{svn}->checkout(
                    $logrepo_url,
                    $logrepo_dir,
                    'HEAD',
                    0, # no recursion
                );
            };
            if ($@) {
                croak "Can't checkout $logrepo_url: $@\n";
            }

            my $logrepo_pkgdir = $logrepo_dir . "/$package";
            mkdir $logrepo_pkgdir;

            open(my $out, ">", "$logrepo_pkgdir/log")
                or croak "Can't open $logrepo_pkgdir/log for writing: $!\n";
            print $out $changelog;
            close($out);

            $self->{svn}->add($logrepo_pkgdir, 1);

            my $message = $options{message} || "import $package";

            $self->commit_directory(
                $logrepo_pkgdir,
                %options,
                package => $package,
                message => $message
            );
        } else {
            open(my $out, ">", "$txtrepo_pkgdir/old_changelog.txt")
                or croak "Can't open $txtrepo_pkgdir/old_changelog.txt for writing: $!\n";
            print $out $changelog;
            close($out);
        }
    }

    # manage sources separatly, if binrepos is used
    if ($self->{binrepos}) {
        my $binrepo_url = $self->get_packages_top_url(%options, binary => 1);
        my $binrepo_dir = $dir . '/binrepo';
        eval {
            $self->{svn}->checkout(
                $binrepo_url,
                $binrepo_dir,
                'HEAD',
                0, # no recursion
            );
        };
        if ($@) {
            croak "Can't checkout $binrepo_url: $@\n";
        }

        my $binrepo_pkgdir = $binrepo_dir . "/$package";
        mkdir $binrepo_pkgdir;
        mkdir $binrepo_pkgdir . "/current";
        mkdir $binrepo_pkgdir . "/current/SOURCES";

        # split sources
        $self->split_sources(
            "$txtrepo_pkgdir/current/SOURCES",
            "$binrepo_pkgdir/current/SOURCES",
        );

        $self->{svn}->add($binrepo_pkgdir, 1);

        my $message = $options{message} || "import $package";

        $self->commit_directory(
            $binrepo_pkgdir,
            %options,
            package => $package,
            message => $message
        );
    }

    $self->{svn}->add($txtrepo_pkgdir, 1);

    my $message = $options{message} || "import $package";

    $self->commit_directory(
        $txtrepo_dir,
        %options,
        package => $package,
        message => $message
    );

    $self->submit_package($package, %options) if $options{'submit'};

    # remove temp directory immediatly, not at program termination
    rmtree($dir);

}

=head2 commit_directory($dir, %options)

Commit a directory into the repository. Return revision number.

Available options:

=over

=item message => $message

Use given commit message  (default: launch interactive editor).

=back

=cut

sub commit_directory {
    my ($self, $dir, %options) = @_;

    return unless $self->{commit};

    my $package = $options{package} || $dir;

    if ($options{sync}) {
        $self->split_sources("$dir/SOURCES", "$dir/SOURCES-bin", link => 1);
        $self->sync_sources(
            $self->find_unsync_sources(
                $dir, $self->find_current_sources($dir)
            )
        );
    }

    $self->_print_msg(3, "Committing directory %s", $package);

    # set log callback
    my $message = $options{message};
    if (!$message) {

        my $changes = '';
        my $find_changes = sub {
            my ($entry, $status) = @_;
            # skip directory entries
            return if -d $entry;

            my $text_status = $status->text_status();
            my $change =
                $text_status == $SVN::Wc::Status::modified ? "modified" :
                $text_status == $SVN::Wc::Status::added    ? "added"    :
                $text_status == $SVN::Wc::Status::deleted  ? "removed"  :
                                                             ""         ;
            $changes .= sprintf(
                "file %s %s\n", File::Spec->abs2rel($entry), $change
            ) if $change;
        };

        foreach my $subdir (qw/SPECS SOURCES/) {
            $self->{svn}->status(
                "$dir/$subdir",
                'HEAD',
                $find_changes,
                1, # recursive
                1, # get_all
                0, # update
                1, # no_ignore
            );
        }

        $message = $self->_get_commit_message_from_editor($changes);
    }
    $self->{svn}->log_msg(
        sub {
            $_[0] = \$message;
            return 0;
        }
    );

    # switch to actual repository if needed
    if ($self->{mirror}) {
        $self->{svn}->switch(
            $dir, 
            $self->get_package_url($package, write => 1),
            'HEAD',
            1,
        );
    }

    my $info = $self->{svn}->commit($dir, 0);
    my $revision = $info ? $info->revision() : undef;

    # reset log callback
    $self->{svn}->log_msg(undef);

    # switch back to mirror if needed
    if ($self->{mirror}) {
        $self->{svn}->switch(
            $dir, 
            $self->get_package_url($package),
            'HEAD',
            1,
        );
    }

    return $revision;
}


=head2 tag_package($package, %options)

Tag a package into the repository directly.

Available options:

=over

=item revision => $revision

Work on a specific revision.

=item message => $message

Use given commit message (default: "tag EV-R", where E is epoch, V is version,
and R is release).

=back

=cut

sub tag_package {
    my ($self, $package, %options) = @_;

    $self->_print_msg(3, 'Tagging package %s', $package);

    my $file = File::Temp->new(CLEANUP => 1);
    my $url = $self->get_package_url($package) . "/SPECS/$package.spec";

    eval {
        $self->{svn}->cat(
            $file,
            $url,
            $options{revision} || DEFAULT_REVISION,
        );
    };
    if ($@) {
        croak "Can't get spec file from $url: $@\n";
    }

    close($file);

    my $spec = RPM::Spec->new($file, anyarch => 1, force => 1)
        or croak "Can't parse spec file from $url\n";
    my $header = $spec->srcheader;

    my $ev = $header->queryformat('%|EPOCH?{%{EPOCH}:}:{}|%{VERSION}');
    my $re = $header->queryformat('%{RELEASE}');

    my $package_url = $self->get_package_url($package, write => 1);
    my $tag_url = $self->get_package_url(
        $package, version => 'releases', write => 1
    );
    my $pristine_url = $self->get_package_url(
        $package, version => 'pristine', write => 1
    );

    $self->{svn}->mkdir($tag_url)
        if !$self->_check_url_exists($tag_url);

    $self->{svn}->mkdir("$tag_url/$ev")
        if !$self->_check_url_exists("$tag_url/$ev");

    croak "$tag_url/$ev/$re already exists\n"
        if $self->_check_url_exists("$tag_url/$ev/$re");

    my $message = $options{message} || "tag release $ev-$re";
    $self->{svn}->log_msg(
        sub {
            $_[0] = \$message;
            return 0;
        }
    );

    $self->_print_msg(
        4, 'Tagging package %s to %s/%s in text repository',
        $package, $ev, $re
    );
    $self->{svn}->copy(
        $package_url,
        $options{revision} || DEFAULT_REVISION,
        "$tag_url/$ev/$re",
    );

    $self->_print_msg(
        4, 'Tagging package %s to pristine in text repository', $package
    );
    eval {
        $self->{svn}->delete($pristine_url, 1);
    };
    $self->{svn}->copy(
        $package_url,
        $options{revision} || DEFAULT_REVISION,
        $pristine_url
    );
    if ($self->{binrepos}) {
        $package_url = $self->get_package_url(
            $package, write => 1, binary => 1
        );
        $tag_url = $self->get_package_url(
            $package, version => 'releases', write => 1, binary => 1
        );
        $pristine_url = $self->get_package_url(
            $package, version => 'pristine', write => 1, binary => 1
        );

        $self->{svn}->mkdir($tag_url)
            if !$self->_check_url_exists($tag_url);

        $self->{svn}->mkdir("$tag_url/$ev")
            if !$self->_check_url_exists("$tag_url/$ev");

        croak "$tag_url/$ev/$re already exists\n"
            if $self->_check_url_exists("$tag_url/$ev/$re");

        $self->_print_msg(
            4,
            'Tagging package %s to %s/%s in binary repository',
            $package, $ev, $re
        );
        $self->{svn}->copy(
            $package_url,
            $options{revision} || DEFAULT_REVISION,
            "$tag_url/$ev/$re",
        );

        $self->_print_msg(
            4,
            'Tagging package %s to pristine in binary repository',
            $package
        );
        eval {
            $self->{svn}->delete($pristine_url, 1);
        };
        $self->{svn}->copy(
            $package_url,
            $options{revision} || DEFAULT_REVISION,
            $pristine_url
        );
    }
    $self->{svn}->log_msg(undef);
}

=head2 get_package_info($package, %options)

Return an hashref containing useful information about given package:

=over

=item size

The size of the package (sum of files size)

=item last_rev

The revision of the last changed

=item last_author

The author of the last change

=item last_time

The time of last change

=back

Available options:

=over

=item revision => $revision

Work on a specific revision.

=item distribution => $distribution

Work on a specific distribution tree.

=item branch => $branch

Work on a specific package branch inside a distribution.

=back

=cut

sub get_package_info {
    my ($self, $package, %options) = @_;
    my $url = $self->get_package_url($package, %options);
    my $leaves;

    eval {
        $leaves = $self->{svn}->ls(
            $url, 
            $options{revision} || DEFAULT_REVISION,
            1,
        );
    };
    if ($@) {
        if ($@ =~ /^Filesystem has no item/) {
            croak "No such package $package\n";
        } else {
            croak "Can't get information from $url: $@\n";
        }
    }
    my $info = {
        size        => 0,
        last_rev    => 0,
        last_time   => 0,
        last_author => ''
    };
    foreach my $leaf (%{$leaves || {}}) {
        next unless defined $leaves->{$leaf};
        $info->{size} += $leaves->{$leaf}->size;
        if ($leaves->{$leaf}->created_rev > $info->{last_rev}) {
            $info->{last_rev} = $leaves->{$leaf}->created_rev();
            $info->{last_time} = $leaves->{$leaf}->time();
            $info->{last_author} = $leaves->{$leaf}->last_author();
        }
    }

    return $info;
}

=head2 rename_package($old_package, $new_package, %options)
=cut

sub rename_package {
    my ($self, $old_package, $new_package, %options) = @_;
}

=head2 submit_package($package, %options)

Submit the package on the build host.

Available options:

=over

=item revision => $revision

Work on a specific revision (default to last revision).

=item distribution => $distribution

Work on a specific distribution tree.

=item branch => $branch

Work on a specific package branch inside a distribution.

=item target => $target

Use given target as submission target (default to repsys configuration value).

=item define => values

Pass each given value to underlying command, using --define switch.

=back

=cut

sub submit_package {
    my ($self, $package, %options) = @_;

    return unless $self->{submit};

    $self->_print_msg(3, 'Submitting package %s', $package);
    
    my $parent_url = $self->get_package_parent_url($package, %options);
    if (!$self->_check_url_exists($parent_url, %options)) {
        croak "$package is not in svn\n";
    }

    my $revision = $options{revision} ||
                   $self->get_package_info($package, %options)->{last_rev};
    my $target   = $options{target} || 
                   $self->{config}->val('submit', 'default');
    my $host       = $self->{config}->val('submit', 'host');
    my $createsrpm = $self->{config}->val('helper', 'create_srpm');
    
    # TODO we can also use xml-rpc, even if not implemented on the server side
    my @command = (
        'ssh',
        '-n',
        $host,
        $createsrpm,
        $parent_url,
        '-r', $revision, 
        '-t', $target,
        (map { ('--define' => $_) } @{$options{define} || []}),
    );

    my $result = system(@command);
    if ($result != 0) {
        croak("Error during submission: $?\n");
    }
}

=head2 update_package($package, $version, %options)

Update a package to a new version.

Available options:

=over

=item revision => $revision

Work on a specific revision.

=item distribution => $distribution

Work on a specific distribution tree.

=item branch => $branch

Work on a specific package branch inside a distribution.

=item force-version => $version

Force given package version.

=item force-release => $release

Force given package release.

=item keep-on-failure => true/false

Keep temporary directory in case of failure

=item message => $message

Use given commit message (default: "update to new version X", where X is new
version if given, "rebuild" otherwise)

=item target => $target

Use given target as submission target.

=item nobuild => true/false

Skip local build of the package.

=back

=cut

sub update_package {
    my ($self, $package, $version, %options) = @_;

    $self->_print_msg(2,
        'Updating package %s' . ($version ? ' to %s' : ''), $package, $version
    );

    my $dir = $options{tempdir} || tempdir(
        TEMPDIR_TEMPLATE,
        TMPDIR => 1,
        CLEANUP => ($options{'keep-on-failure'} ? 0 : 1)
    );

    $self->checkout_package(
        $package,
        $dir,
        %options
    );

    # ensure a full second delay between checkout and spec update
    # otherwise svn doesn't see differences
    sleep(1);

    $self->_set_rpm_dirs_from_topdir($dir);

    my $spec_file =
        $options{specfile} ||
        (glob(RPM::expand_macro('%_specdir/*.spec')))[0];

    Youri::Package::RPM::Updater->require(qv('0.4.0')) or croak $@;
    my $updater = Youri::Package::RPM::Updater->new(
        sourcedir => "$dir/SOURCES",
        verbose                =>
            ($self->{verbosity} > 2 ? $self->{verbosity} - 2 : 0)
    );

    eval {
        $updater->update_from_spec(
            $spec_file,
            $version,
            ($options{'release'} ?
                (release => $options{'release'}) :
                ()
            ),
            ($options{'spec-line-expression'} ?
                (spec_line_expression => $options{'spec-line-expression'}) :
                ()
            ),
            update_changelog => 0,
            # don't check version for perl packages
            check_new_version => ($package =~ /^perl-/ ? 0 : 1)
        );
    }; 
    if ($@) {
        croak "Can't update package: $@\n";
    }

    $self->build_package($spec_file, %options);

    # check current svn status before syncing
    $self->split_sources("$dir/SOURCES", "$dir/SOURCES-bin", link => 1);
    $self->sync_sources(
        $self->find_unsync_sources(
            $dir, $self->find_current_sources($dir)
        )
    );

    my $message = $options{message} ? $options{message}
                : $version          ? "update to new version $version"
                :                     "rebuild" 
                ;
    $self->commit_directory($dir, message => $message);

    $self->submit_package($package, %options);

    # remove temp directory immediatly, not at program termination
    rmtree($dir);
}

=head2 mass_update($url)

Use given URL as a list of available packages update, and call update package
foreach of them

Available options:

=over

=item include => $regexp

Only process package matching given regexp

=item exclude => $regexp

Don't process package matching given regexp

=item keep-on-failure => true/false

Keep temporary directories in case of failure

=back

=cut

sub mass_update {
    my ($self, $url, %options) = @_;
    croak 'No URL given' unless $url;

    LWP::UserAgent->require() or croak $@;
    HTML::TableExtract->require() or croak $@;

    $self->_print_msg(1, 'Mass updating from %s', $url);

    $self->_print_msg(2, 'Retrieving report');
    my $agent = LWP::UserAgent->new();
    my $response = $agent->get($url);

    croak 'Unreachable URL: ' . $response->code()
        unless $response->is_success();

    my $parser;
    my $content = $response->content();
    if ($content =~ /<title>age (global|individual) report/) {
        $parser = HTML::TableExtract->new(
            headers => [
                'source package',
            ]
        );
    } elsif ($content =~ /<title>updates (global|individual) report/) {
        $parser = HTML::TableExtract->new(
            headers => [
                'source package',
                'available'
            ]
        );
    } else {
        croak 'Unknown report type';
    }

    $self->_print_msg(2, 'Parsing report content');
    $parser->parse($content);
    my ($include, $exclude);
    $include = qr/$options{include}/ if $options{include};
    $exclude = qr/$options{exclude}/ if $options{exclude};

    my $dir = tempdir(TEMPDIR_TEMPLATE, TMPDIR => 1, CLEANUP => 0);
    my $error_occured = 0;
    foreach my $row ($parser->rows()) {
        my ($name, $version) = ($row->[0], $row->[1]);
        # skip additional lines for spanning rows
        next unless $name; 
        next if $exclude && $name =~ $exclude;
        next if $include && $name !~ $include;
        eval {
            $self->update_package(
                $name,
                $version,
                tempdir => "$dir/$name"
            );
        };
        if ($@) {
            warn "Update failed for $name" .
                ($version ? " $version" : "") .
                "\n";
            $error_occured = 1;
        }
    }

    # remove temp directory immediatly, not at program termination
    rmtree($dir) unless $options{'keep-on-failure'} && $error_occured;
}

# private methods

# Set rpm macros corresponding to build directories, either relatively to given
# top directory if relative path is given, or absolutly if absolute path is
# given.

sub _set_rpm_dirs_from_topdir {
    my ($self, $topdir, %other_dirs) = @_;

    $topdir = File::Spec->rel2abs($topdir);

    foreach my $name (keys %rpm_dirs) {
        my $path = $other_dirs{$name} || $rpm_dirs{$name};
        if ($path !~ /^\//) {
            $path = "$topdir/$path";
        }
        RPM::add_macro("_$name $path");
    }
}

# Set rpm macros corresponding to build directories, calling
# _set_rpm_dirs_from_topdir with the parent directory of given spec file
# directory.

sub _set_rpm_dirs_from_specfile {
    my ($self, $specfile, %other_dirs) = @_;

    my ($topdir) = $specfile =~ m{^(.*)/.*/.*$};
    $self->_set_rpm_dirs_from_topdir($topdir, %other_dirs);
}

# Create rpm build directories

sub _create_rpm_dirs {
    my ($self, @dirs) = @_;

    @dirs = keys %rpm_dirs unless @dirs;

    foreach my $name (@dirs) {
        my $path = RPM::expand_macro("%_$name");
        if (! -d $path) {
            $self->_print_msg(5, 'Creating directory %s', $path);
            if (!mkdir($path)) {
                croak "can't create directory $path: $!\n";
            }
        }
    }
}

# Print a trace message, according to current log level

sub _print_msg {
    my ($self, $level, $fmt, @args) = @_;
    return if $level > $self->{verbosity};
    printf("$fmt\n", @args);
}

# Get a commit message using an external editor

sub _get_commit_message_from_editor {
    my ($self, $footer) = @_;

    my $delimiter = "--This line, and following ones, will be ignored--";
    
    my $tmp = File::Temp->new(CLEANUP => 1);
    print $tmp "\n";
    print $tmp $delimiter . "\n";
    print $tmp $footer if $footer;
    close($tmp);

    my ($editor) =
        grep { $_ }
        map { $ENV{$_} }
        qw(SVN_EDITOR VISUAL EDITOR);
    $editor ||= 'vi';

    if (system($editor, $tmp->filename()) == -1) {
        croak "Can't start $editor\n";
    }

    open(my $in, "<", $tmp->filename())
        or croak "Can't read editor output\n";

    my $message = '';
    while (my $line = <$in>) {
        last if $line =~ /^$delimiter$/;
        $message .= $line;
    }
    close($in);
    chomp($message);

    croak "empty commit message, aborting" unless $message;

    return $message;
}

=head2 get_old_changelog_raw($package, %options)

Get old changelog for package, in raw format.

Available options:

=over

=item distribution => $distribution

Work on a specific distribution tree.

=item branch => $branch

Work on a specific package branch inside a distribution.

=back

=cut

sub get_old_changelog_raw {
    my ($self, $package, %options) = @_;

    my $changelog;
    open(my $out, '>', \$changelog) or croak "Can't open  in-memory file: $!";

    my $url;
    my $split_changelog = $self->{config}->val('repository', 'split_changelog');
    if ($split_changelog && $split_changelog eq 'yes') {
        $url =
            $self->{config}->val('repository', 'main') .
            "/misc/$package/log";
    } else {
        $url = $self->get_package_parent_url($package, %options) .
            "/old_changelog.txt";
    }

    eval {
        $self->{svn}->cat(
            $out,
            $url,
            $options{revision} || DEFAULT_REVISION,
        ); 
    };
    if ($@) {
        if ($@ =~ /^Filesystem has no item/) {
            # no changelog for this package
        } else {
            croak "Can't get old changelog for $package: $@\n";
        }
    }
    close($out);

    return $changelog;
}

=head2 get_old_changelog($package, %options)

Get old changelog for package, in structured format.

Available options:

=over

=item distribution => $distribution

Work on a specific distribution tree.

=item branch => $branch

Work on a specific package branch inside a distribution.

=back

=cut

sub get_old_changelog {
    my ($self, $package, %options) = @_;

    my $changelog = $self->get_old_changelog_raw($package, %options);

    my @changelog;

    foreach my $line (split(/\n/, $changelog)) {
        chomp($line);
        $line or next;
        $line =~ /^%changelog/ and next;
        if ($line =~ /^\* (\w+\s+\w+\s+\d+\s+\d+)\s+(.*)/) {
            push(
                @changelog,
                {
                    time   => str2time($1, 'UTC') || 0,
                    author => $2 || '',
                    text   => '',
                }
            );
        } else {
            if (!@changelog) {
                push(@changelog, { header => 1, text => '', 'time' => 0 });
            }
            $changelog[-1]->{text} .= "$line\n";
        }
    }

    return @changelog;
}

=head2 get_new_changelog($package, %options)

Get new changelog for package, in structured format.

Available options:

=over

=item distribution => $distribution

Work on a specific distribution tree.

=item branch => $branch

Work on a specific package branch inside a distribution.

=back

=cut

sub get_new_changelog {
    my ($self, $package, %options) = @_;

    my @entries;

    my $release_callback = sub {
        my ($changed_paths, $revision, $author, $date, $message) = @_;
        foreach my $path (keys %{$changed_paths || {}}) {
            my $info = $changed_paths->{$path};
            my $from_revision = $info->copyfrom_rev();
            next unless $from_revision > 0;
            next unless $path =~ m{releases/(?:([^:/]+):)?([^/]+)/([^/]+)};
            push(
                @entries, {
                    revision => $from_revision,
                    author   => '',
                    time     => str2time($date),
                    text     => '',
                    evr      => $1 ? "$1:$2-$3" : "$2-$3",
                }
            );
        }
    };

    my $release_url =
        $self->get_package_url($package, %options, version => 'releases');
    if ($self->_check_url_exists($release_url)) {
        eval {
            $self->{svn}->log(
                $release_url,
                $options{revision} || DEFAULT_REVISION,
                0, 1, 0,
                $release_callback,
            );
        };
        if ($@) {
            croak "Can't get svn log from release branch: $@\n";
        }
    }

    my $current_callback = sub {
        my ($changed_paths, $revision, $author, $date, $message) = @_;
        my $entry_to_update;
        foreach my $entry (@entries) {
            last if $revision > $entry->{revision};
            $entry_to_update = $entry;
        }

        if (!$entry_to_update) {
            $entry_to_update = {
                revision => $revision,
                author   => $author,
                time     => str2time($date),
                text     => '',
            };

            push(@entries, $entry_to_update);
        }

        $entry_to_update->{author} ||= $author;
        my @gti = gmtime(str2time($date));

        my ($textentry) =
            grep { $_->{author} eq $author }
            @{$entry_to_update->{changesets} || []};

        if (!$textentry) {
            $textentry = {
                author => $author,
                text => [],
            };
            push(@{$entry_to_update->{changesets}}, $textentry);
        }

        push(@{$textentry->{text}}, $message);
        
    };

    eval {
        $self->{svn}->log(
            $self->get_package_url($package, %options),
            $options{revision} || DEFAULT_REVISION,
            0, 0, 0,
            $current_callback,
        );
    };
    if ($@) {
        croak "Can't get svn log from current branch: $@\n";
    }

    return @entries;
}

# format a changelog entry

sub _format_changelog_entry {
    my ($self, $entry) = @_;
    $entry->{time} or return $entry->{text};
    my @gti = gmtime($entry->{time});

    # subversion changelog is having empty commit
    return if not $entry->{author};
    my $text = $entry->{text};
    if (!$text) {
        my $indent = '';
        foreach my $changeset (@{$entry->{changesets} || []}) {
            if ($changeset->{author} ne $entry->{author}) {
                $text .=
                    "- from " .
                    $self->_get_formated_author($changeset->{author}) .
                    "\n";
            }
            foreach my $change (@{$changeset->{text}}) {
                next if $change =~ /SILENT/;
                my $dash = '- ';
                $change =~ s/^(\s|-|\*)*/$indent- /;
                foreach my $line (split(/\n/, $change)) {
                    next if $line =~ /SILENT/;
                    $line =~ s/^(\s|\*)*/$indent/;
                    if ($line !~ /^-/) {
                        $line =~ s/^/  /;
                    }
                    $text .= $line . "\n";
                    $dash = '  ';
                }
            }
            $indent = '  ';
        }
    }
    $text =~ s/^\*/-/gm;

    my $formated_entry;
    $formated_entry .= sprintf "* %s %s %s\n",
        strftime("%a %b %d %Y", @gti),
        $self->_get_formated_author($entry->{author}),
        $entry->{evr} || '';
    $formated_entry .= sprintf "+ Revision: %s\n", $entry->{revision}
        if $entry->{revision};
    $formated_entry .= sprintf "%s\n", $text;

    return $formated_entry;
}

# check given url exists

sub _check_url_exists {
    my ($self, $url, %options) = @_;

    eval {
        $self->{svn}->ls(
            $url, 
            $options{revision} || DEFAULT_REVISION,
            0,
        );
    };
    if ($@) {
        return 0;
    }

    return 1;
}

# check given source exists

sub _check_source_exists {
    my ($self, $source) = @_;

    my $text_status;
    $self->{svn}->status(
        $source,
        'HEAD',
        sub {
            my (undef, $status) = @_;
            $text_status = $status->text_status();
        },
        1, # recursive
        1, # get_all
        0, # update
        1, # no_ignore
    );

    return $text_status != $SVN::Wc::Status::unversioned;
}

sub _get_formated_author {
    my ($self, $author) = @_;

    # return cached result if it exists
    return $self->{authors}->{$author} if $self->{authors}->{$author};

    my $formated_author;
    my $server = $self->{config}->val('ldap', 'server');
    if ($server) {
        # lazy ldap connection
        if (!$self->{ldap}) {
            Net::LDAP->require();
            $self->{ldap} = Net::LDAP->new($server);
            $self->{ldap}->bind();
        }

        my $base = $self->{config}->val('ldap', 'base'),
        my $filter = $self->{config}->val('ldap', 'filter_format');
        $filter =~ s/\$username/$author/;

        my $result = $self->{ldap}->search(base => $base, filter => $filter);

        if ($result->count() == 0) { 
            $formated_author = $author;
        } else {
            my $entry = $result->entry(0);
            $formated_author = $self->{config}->val('ldap', 'result_format');
            $formated_author =~ s/\$(\w+)/$entry->get_value($1)/eg;
        }
    } else {
        $formated_author = $self->{config}->val('users', $author, $author);
    }

    $self->{authors}->{$author} = $formated_author;
    return $formated_author;
}

1;
